"""Console adapted from Python's console found in code.py.

An earlier version of this code was subclassing code.InteractiveConsole.
However, as more and more changes were introduced, dealing with
code transformation and especially customized error handling,
it seemed to make sense to "rewrite" every relevant part in
this single module.
"""
import os
import platform
import sys

from codeop import CommandCompiler
from tokenize import TokenError

from . import version
from .converter import convert
from .session import state
from .exception_handling import write_exception_info
from .exceptions import AvantPyException


class AvantPyInteractiveConsole:
    """A Python console that tries to emulate the normal Python interpreter
       except that it support experimental code transformations.
       It is adapted from cPython's ``code.InteractiveConsole`` and its
       parent.

       Like the normal Python interactive console, it attempts to evaluate
       code entered one line at a time by a user.
    """

    def __init__(self, locals=None, show_python=False):
        self.show_python = show_python
        self.locals = locals if locals is not None else {}
        self.compile = CommandCompiler()
        self.name = "<AvantPy console>"
        self.reset_buffer()
        state.console_active = True

    def interact(self, banner=None):
        """Emulates the interactive Python console.
        """
        self.write("%s\n" % str(banner))
        more = False
        while True:
            try:
                if more:
                    prompt = state.prompt2
                else:
                    prompt = state.prompt1
                try:
                    line = input(prompt)
                except EOFError:
                    self.write("\n")
                    break
                else:
                    more = self.push(line)
            except KeyboardInterrupt:
                self.write("\nKeyboardInterrupt\n")
                self.reset_buffer()
                more = False

    def push(self, line):
        """Pushes a transformed line to the interpreter.

        The line should not have a trailing newline; it may have
        internal newlines.  The line is transformed and appended to a buffer.
        The interpreter's run_source() method is called with the
        concatenated contents of the buffer as source.  If this
        indicates that the command was executed or invalid, the buffer
        is reset; otherwise, the command is incomplete, and the buffer
        is left as it was after the line was appended.  The return
        value is True if more input is required, False if the line was dealt
        with in some way (this is the same as run_source()).
        """
        assert not line.endswith("\n"), "Forbidden trailing newline in push()."

        self.buffer.append(line)
        self.source = "\n".join(self.buffer)

        try:
            self.converted = convert(self.source, source_name=self.name)
            self.identical = self.converted == self.source
        except SystemExit:
            os._exit(1)
        except AvantPyException as exc:
            write_exception_info(exc, self.source)
            self.reset_buffer()
            return False
        except TokenError as exc:
            exc.args[0].startswith("EOF")
            return True
        except Exception as exc:
            print(_("UNHANDLED EXCEPTION in console.py. This should not happen."))
            raise exc

        try:
            more = self.run_source(self.converted)
        except SystemExit:
            os._exit(1)
        except Exception as exc:
            write_exception_info(exc, self.source)
            self.reset_buffer()
            return False

        if not more:
            self.reset_buffer()
            if self.show_python and not self.identical:
                self.show_converted()
        return more

    def reset_buffer(self):
        """Reset the input buffer."""
        self.buffer = []

    def run_source(self, source, symbol="single"):
        """Compile and run some source in the interpreter.

        Arguments are as for compile_command().

        One several things can happen:

        1) The input is incorrect; compile_command() raised an
        exception (SyntaxError or OverflowError).

        2) The input is incomplete, and more input is required;
        compile_command() returned None.  Nothing happens.

        3) The input is complete; compile_command() returned a code
        object.  The code is executed by calling self.runcode() (which
        also handles run-time exceptions, except for SystemExit).

        The return value is True in case 2, False in the other cases (unless
        an exception is raised).  The return value can be used to
        decide which prompt to use next line.
        """

        try:
            code = self.compile(source, self.name, symbol)
        except (OverflowError, SyntaxError, ValueError) as exc:
            # Case 1
            if self.show_python and not self.identical:
                self.show_converted(self.converted)
            write_exception_info(exc, source)
            return False

        if code is None:  # Case 2
            return True

        # Case 3
        self.runcode(code)
        return False

    def runcode(self, code):
        """Execute a code object.

        All exceptions are caught except SystemExit, which is reraised.

        A note about KeyboardInterrupt: this exception may occur
        elsewhere in this code, and may not always be caught.  The
        caller should be prepared to deal with it.
        """
        try:
            exec(code, self.locals)
        except SystemExit:
            raise
        except Exception as exc:
            write_exception_info(exc, self.source)

    def show_converted(self):
        """Prints the converted source"""
        print()
        for line in self.converted.split("\n"):
            print("|", line)
        print()
        self.identical = True  # prevent from showing again

    def write(self, data):
        """Write a string.

        The base implementation writes to sys.stderr; a subclass may
        replace this with a different implementation.

        """
        sys.stderr.write(data)


def start_console(local_vars=None, show_python=False):
    """Starts a console; modified from code.interact"""
    console_defaults = {"set_lang": state.set_lang, "set_dialect": state.set_dialect}

    if local_vars is None:
        local_vars = console_defaults
    else:
        local_vars.update(console_defaults)

    console = AvantPyInteractiveConsole(locals=local_vars, show_python=show_python)

    banner = "AvantPy version {}. [Python version: {}]\n".format(
        version.__version__, platform.python_version()
    )
    console.interact(banner=banner)



"""French dialect"""

pyfr = {
    "False": "Faux",
    "None": "Aucun",
    "True": "Vrai",
    "and": "et",
    "as": "comme",
    "assert": "affirmer",
    "async": "async",  # do not translate
    "await": "await",  # as these are not for beginners
    "break": "interrompre",
    "class": "classe",
    "continue": "continuer",
    "def": "définir",
    "del": "supprimer",
    "elif": "sinonsi",
    "else": ["sinon", "pasinterrompu"],
    "except": "siexception",
    "finally": "finalement",
    "for": "pour",
    "from": "de",
    "global": "global",
    "if": "si",
    "import": "importer",
    "in": "dans",
    "is": "est",
    "lambda": "fonction",
    "nonlocal": "nonlocal",
    "not": "pas",
    "or": "ou",
    "pass": "passer",
    "raise": "lever",
    "return": "retourner",
    "try": "essayer",
    "while": "tantque",
    "with": "avec",
    "yield": "cède",
    # Special loop keywords
    "repeat": "répéter",
    "forever": "sansfin",
    "until": "jusquà",
    # a few builtins useful for beginners
    "input": "demander",
    "print": "afficher",
    "range": "intervalle",
    "exit": "quitter",  # useful for console
    # a well-known python expression
    '__name__ == "__main__"': "pasimporté",
}



"""This module provides a GUI to the conversion function.

It makes it possible to load up a source file written in a given dialect
and convert it into another dialect or into Python, showing both the
original source and the converted one in two different editor frames,
with Python keyword, as well as their counterpart in various dialects,
highlighted.
"""
import os
import tkinter as tk
import tokenize

from io import StringIO
from tkinter import filedialog, ttk

from .converter import transcode
from .session import state
from .utils import Token
from . import exception_handling


class EditorWidget(tk.Frame):
    """A scrollable text editor, that can save files."""

    def __init__(self, parent, parent_frame, other=None):
        super().__init__()
        self.parent = parent
        self.parent_frame = parent_frame
        self.other = other
        self.frame = tk.Frame(self, width=600, height=600)
        self.text_area = self.init_text_area()
        self.set_horizontal_scroll()
        self.set_vertical_scroll()
        self.linenumbers = TextLineNumbers(self.frame, width=30)
        self.linenumbers.attach(self.text_area)
        if self.other is None:
            self.linenumbers.pack(side="right", fill="y")
        else:
            self.linenumbers.pack(side="left", fill="y")
        self.text_area.pack(side="left", fill="both", expand=True)
        self.text_area.bind("<Key>", self.colorize)
        self.frame.pack(fill="both", expand=True)

    def init_text_area(self):
        text_area = tk.Text(
            self.frame,
            wrap="none",
            width=600,
            height=600,
            padx=10,
            pady=10,
            font="Monaco 12",
        )
        text_area.tag_config("Python", font="Monaco 12 bold", foreground="forest green")
        text_area.tag_config("source", font="Monaco 12 bold", foreground="blue")
        text_area.tag_config(
            "converted", font="Monaco 12 bold", foreground="dark violet"
        )
        return text_area

    def set_horizontal_scroll(self):
        scroll_x = tk.Scrollbar(
            self.frame, orient="horizontal", command=self.text_area.xview
        )
        scroll_x.config(command=self.text_area.xview)
        self.text_area.configure(xscrollcommand=scroll_x.set)
        scroll_x.pack(side="bottom", fill="x", anchor="w")

    def set_vertical_scroll(self):
        # Scroll Bar y For Height
        def scroll_vertical(*args):
            self.text_area.yview(*args)
            if self.other is not None:
                self.other.text_area.yview(*args)

        if self.other is not None:
            scroll_y = tk.Scrollbar(self.frame)
            scroll_y.config(command=scroll_vertical)
            self.text_area.configure(yscrollcommand=scroll_y.set)
            scroll_y.pack(side="right", fill="y")

    def insert_text(self, txt):
        """Inserts the text in the editor, replacing any previously existing
           content.
        """
        self.text_area.delete("1.0", tk.END)
        self.text_area.insert("1.0", txt)
        self.parent.update_idletasks()
        self.colorize()

    def get_text(self):
        """Gets the current content and returns it as a string"""
        # For some reason, an extra "\n" is added, which we need to remove.
        return self.text_area.get("1.0", tk.END)[:-1]

    def colorize(self, event=None):
        """Colorizes the Python keywords and a few builtins, as well as the
           corresponding version in other dialects.
        """
        content = self.text_area.get("1.0", tk.END)
        try:
            tokens = tokenize.generate_tokens(StringIO(content).readline)
            for tok in tokens:
                token = Token(tok)
                if token.string in self.parent_frame.python_words:
                    begin_index = "{0}.{1}".format(token.start_line, token.start_col)
                    end_index = "{0}.{1}".format(token.end_line, token.end_col)
                    self.text_area.tag_add("Python", begin_index, end_index)
                elif token.string in self.parent_frame.source_words:
                    begin_index = "{0}.{1}".format(token.start_line, token.start_col)
                    end_index = "{0}.{1}".format(token.end_line, token.end_col)
                    self.text_area.tag_add("source", begin_index, end_index)
                elif token.string in self.parent_frame.converted_words:
                    begin_index = "{0}.{1}".format(token.start_line, token.start_col)
                    end_index = "{0}.{1}".format(token.end_line, token.end_col)
                    self.text_area.tag_add("converted", begin_index, end_index)
        except tokenize.TokenError:
            pass

    def save_file(self, event=None):
        """Saves the file currently in the Texteditor"""
        filename = filedialog.asksaveasfilename(
            filetypes=(("AvantPy/Python", "*.py*"),)
        )
        if filename is not None:
            with open(filename, "w", encoding="utf8") as f:
                data = self.get_text()
                f.write(data)


class TextLineNumbers(tk.Canvas):
    def __init__(self, *args, **kwargs):
        tk.Canvas.__init__(self, *args, **kwargs)
        self.textwidget = None

    def attach(self, textwidget):
        self.textwidget = textwidget
        self.redraw()

    def redraw(self, *args):
        """redraw line numbers"""
        self.delete("all")

        i = self.textwidget.index("@0,0")
        while True:
            dline = self.textwidget.dlineinfo(i)
            if dline is None:
                break
            y = dline[1]
            linenum = str(i).split(".")[0]
            self.create_text(2, y, anchor="nw", text=linenum, font="Monaco 12")
            i = self.textwidget.index("%s+1line" % i)
        self.after(30, self.redraw)


class App(tk.Frame):
    def __init__(self, master=None):
        tk.Frame.__init__(self, master)
        self.grid()
        self.master.title("Converter")
        self.all_dialects = ["py"] + state.all_dialects()
        ext = ["*.%s" % d for d in self.all_dialects]
        self.filetypes = " ".join(ext)
        self.add_ui()
        self.python_words = []
        self.source_words = []
        self.converted_words = []

    def add_ui(self):
        """Adds UI elements to the main window"""

        for r in range(3):
            self.master.rowconfigure(r, weight=1, minsize=20)
        for c in [1, 5]:
            self.master.columnconfigure(c, weight=1, minsize=40)

        tk.Label(self.master, text="Source").grid(row=0, column=1)
        tk.Label(self.master, text="Converted").grid(row=0, column=4)

        self.converted_editor = EditorWidget(self.master, self)
        self.converted_editor.grid(row=1, column=3, columnspan=4, sticky="news")
        self.source_editor = EditorWidget(
            self.master, self, other=self.converted_editor
        )
        self.source_editor.grid(row=1, column=0, columnspan=3, sticky="news")

        tk.Button(self.master, text="Open", command=self.get_source).grid(
            row=2, column=0, sticky="w"
        )
        tk.Button(self.master, text="Save", command=self.source_editor.save_file).grid(
            row=2, column=1, sticky="w"
        )
        tk.Button(self.master, text="Convert source", command=self.convert_source).grid(
            row=2, column=3, sticky="w"
        )
        # if the following is assigned a local value, it will be garbage collected
        # and the combobox will show nothing initially.
        self._choice = tk.StringVar()
        self.dialects = ttk.Combobox(
            self.master, textvariable=self._choice, values=self.all_dialects
        )
        self.dialects.grid(row=2, column=4, sticky="w")
        self.dialects.current(0)
        self.dialects.bind("<<ComboboxSelected>>", self.get_conversion_dialect)
        self.conversion_dialect = self.all_dialects[0]
        tk.Button(
            self.master, text="Save", command=self.converted_editor.save_file
        ).grid(row=2, column=5, sticky="w")

    def get_conversion_dialect(self, event):
        """Gets the conversion dialect selected from the ComboBox."""
        self.conversion_dialect = self.dialects.get()

    def get_source(self, event=None):
        """Opens a file by looking first from the current directory."""
        txt_file = filedialog.askopenfilename(
            initialdir=os.getcwd(), filetypes=(("AvantPy", self.filetypes),)
        )
        self.source_dialect = txt_file.split(".")[-1]
        if self.source_dialect == "py":
            self.source_dialect = "pyen"
        self.python_words = set(state.get_from_python(self.source_dialect).keys())
        self.source_words = set(state.get_to_python(self.source_dialect).keys())
        if txt_file:
            with open(txt_file, encoding="utf8") as new_file:
                self.source_editor.insert_text(new_file.read())

    def convert_source(self, event=None):
        """Converts the content of the source window into the requested dialect."""
        text = self.source_editor.get_text()
        if self.conversion_dialect != "py":
            self.converted_words = set(
                state.get_to_python(self.conversion_dialect).keys()
            )
        else:
            self.converted_words = set(state.get_from_python("pyen").keys())

        try:
            new_text = transcode(text, self.source_dialect, self.conversion_dialect)
        except Exception as exc:
            exception_handling.write_err = self.converted_editor.insert_text
            exception_handling.write_exception_info(exc, text)
        else:
            self.converted_editor.insert_text(new_text)


def main():
    root = tk.Tk()
    root.geometry("900x600+200+200")
    root.minsize(600, 600)
    app = App(master=root)
    app.mainloop()


if __name__ == "__main__":
    main()



"""utils.py

Classes or functions used in various other files.
"""


class Token:
    """Token as generated from tokenize.generate_tokens written here in
       a more convenient form for our purpose.
    """

    def __init__(self, token):
        self.type = token[0]
        self.string = token[1]
        self.start_line, self.start_col = token[2]
        self.end_line, self.end_col = token[3]
        # ignore last parameter which is the logical line



"""session.py

Keeps track of and updates the state of the session: language and dialect used.
"""
import gettext
import glob
import os.path
import runpy

from . import exceptions


class _State:
    """Keeps track of dialect and lang parameters which control some of the
       behaviour of AvantPy.
    """

    def __init__(self):
        self.dict_to_py = {}
        self.dict_from_py = {}
        self.languages = set([])
        self.current_dialect = None
        self.current_lang = None
        self.current_filename = None
        self.prompt1 = ">>> "  # can be reset based on dialect
        self.prompt2 = "... "
        self.console_active = False
        self.collect_dialects()
        self.collect_languages()
        self.install_gettext("xx")  # Simply adds _() to builtins

    def collect_dialects(self):
        """Find known dialects and create corresponding dictionaries."""
        dialects = glob.glob(os.path.dirname(__file__) + "/dialects/*.py")
        for f in dialects:
            if os.path.isfile(f) and not f.endswith("__init__.py"):

                dialect = os.path.basename(f).split(".")[0]

                _module = runpy.run_path(f)
                from_py = _module[dialect]
                self.dict_from_py[dialect] = from_py

                to_py = {}
                for k, v in from_py.items():
                    if isinstance(v, str):
                        to_py[v] = k
                    else:
                        for val in v:
                            to_py[val] = k
                self.dict_to_py[dialect] = to_py

    def collect_languages(self):
        """Creates a set of known languages."""
        languages = glob.glob(os.path.dirname(__file__) + "/locales/*")
        for f in languages:
            if os.path.isdir(f):
                lang = os.path.basename(f)
                self.languages.add(lang)

    def all_dialects(self):
        """Returns a list of all known dialects."""
        return [k for k in self.dict_from_py.keys()]

    def is_dialect(self, dialect):
        """Returns True if `dialect` is a known dialect, False otherwise."""
        return dialect in self.dict_from_py

    def is_lang(self, lang):
        """Returns True if `lang` is known as part of a dialect,
                   False otherwise
        """
        return lang in self.languages

    def get_dialect(self):
        """Returns the current dialect."""
        return self.current_dialect

    def get_lang(self):
        """Returns the current language."""
        return self.current_lang

    def get_to_python(self, dialect=None):
        """Returns a dict providing translation from a dialect into Python.
        """
        # to_py uses the language code instead of the dialect as key.
        if dialect is None:
            dialect = self.current_dialect
        return self.dict_to_py[dialect]

    def get_from_python(self, dialect=None):
        """Returns a dict providing translation from Python into a dialect."""
        if dialect is None:
            dialect = self.current_dialect
        return self.dict_from_py[dialect]

    def set_dialect(self, dialect):
        """Sets the current dialect.

           Raises an exception if the dialect is unknown.

           If current language is set to None, an attempt will be made
           to set language to the value corresponding to dialect.

           Also sets the default prompt.
        """
        if not self.is_dialect(dialect):
            raise exceptions.UnknownDialectError(
                "Unknown dialect %s; known dialects = %s"
                % (dialect, self.all_dialects()),
                (dialect, self.all_dialects()),
            )
        else:
            lang = dialect[2:]
            self.current_dialect = dialect
            if self.current_lang is None:
                try:
                    self.set_lang(lang)
                except exceptions.UnknownLanguageError:
                    pass
            elif self.console_active:
                self.print_lang_info()
        self.prompt1 = dialect + "> "
        self.prompt2 = "..." + " " * (len(dialect) - 1)

    def set_lang(self, lang):
        """Sets the current language.

           Raises an exception if the language is unknown.

           If current dialect is set to None, an attempt will be made
           to set dialect to the value corresponding to language.
        """
        if not self.is_lang(lang):
            raise exceptions.UnknownLanguageError(
                "Unknown language %s; known languages = %s" % (lang, self.languages),
                (lang, self.languages),
            )
        else:
            self.current_lang = lang
            self.install_gettext(lang)
            if self.current_dialect is None:
                try:
                    self.set_dialect("py" + lang)
                except exceptions.UnknownDialectError:
                    pass
            elif self.console_active:
                self.print_lang_info()

    def install_gettext(self, lang):
        """Sets the current language for gettext."""
        gettext_lang = gettext.translation(
            lang,
            localedir=os.path.normpath(
                os.path.join(os.path.dirname(__file__), "locales")
            ),
            languages=[lang],
            fallback=True,
        )
        gettext_lang.install()

    def print_lang_info(self):
        """Prints language and dialect selected.

           Intended to be used in the console.
        """
        print(
            _("    ==> LANGUAGE: {lang} | AVANTPY DIALECT: {dialect}").format(
                lang=self.current_lang, dialect=self.current_dialect
            )
        )


state = _State()



"""Custom exceptions"""


class AvantPyException(Exception):
    """Base class for custom exceptions."""

    def __init__(self, msg, args):
        super().__init__(msg)
        self.msg = msg
        self.args = args


class IfNobreakError(AvantPyException):
    """Raised if ``nobreak`` is used instead of ``else`` in an ``if`` statement"""

    pass


class MismatchedBracketsError(AvantPyException):
    """Raised if a left parenthesis '(', or square bracket '['
       or curly bracket '{', is closed by a right 'bracket' of a different type.

       In standard Python, this would be indicated as::

           SyntaxError: invalid syntax
    """

    pass


class MissingLeftBracketError(AvantPyException):
    """Raised if a right parenthesis ')', or square bracket ']'
       or curly bracket '}', is found without having a matching left bracket
       found previously.

       In standard Python, this would be indicated as::

           SyntaxError: invalid syntax
    """

    pass


class MissingRepeatError(AvantPyException):
    """Raised if ``until`` or ``forever`` is used without being preceeded by
       ``repeat``.
    """

    pass


class MissingRepeatColonError(AvantPyException):
    """Raised if a line beginning with ``repeat`` does not end with a colon.
    """

    pass


class MissingRightBracketError(AvantPyException):
    """Raised if a left parenthesis '(', or square bracket '['
       or curly bracket '{', is never closed by a corresponding bracket.

       In standard Python, this would be indicated as::

           SyntaxError: unexpected EOF while parsing
    """

    pass


class NobreakFirstError(AvantPyException):
    """Raised if ``nobreak`` is used somewhere other than at the beginning of a line."""

    pass


class NobreakSyntaxError(AvantPyException):
    """Raised if ``nobreak`` is without a matching block."""

    pass


class RepeatFirstError(AvantPyException):
    """Raised if ``repeat`` is used somewhere other than at the beginning of a line."""

    pass


class TryNobreakError(AvantPyException):
    """Raised if ``nobreak`` is used instead of ``else`` in ``try/except`` statement"""

    pass


class UnknownDialectError(AvantPyException):
    """Raised when attempting to set ``dialect`` to unsupported value."""

    pass


class UnknownLanguageError(AvantPyException):
    """Raised when attempting to set ``lang`` to unsupported value."""

    pass



"""The goal of exception_handling.py is to display information about
exceptions being raised in a way that is much easier to understand
for beginners than normal Python tracebacks.  These "friendly" tracebacks
are written so that they can easily be translated into any human language.
"""
import sys
import traceback

from .session import state

ENABLED = True


def disable():
    """Disables custom exception handling"""
    global ENABLED
    ENABLED = False


def get_partial_source(source, begin, end, marks=None, nb_digits=None):
    """Extracts a few relevant lines from a source file.

       If the number of lines would exceed a certain limit (currently 7)
       the function is called recursively to only show the a few lines
       at the beginning and at the end.
    """
    if nb_digits is None:
        nb_digits = len(str(end))
    no_mark = "       {:%d}: " % nb_digits
    with_mark = "    -->{:%d}: " % nb_digits
    continuation = "           ..."

    if marks is None:
        marks = []

    result = []
    if end - begin > 7:
        # Suppose the source spans line numbers 1 to 12. Splitting it to show
        # partial result would show lines 1 to 3, followed by 10 to 12.
        # If we want the indentation of the first part to match the indentation
        # of the second part, we must specify the length of the
        # indentation of the first part.
        # See the traceback for MismatchedBracketsError in the
        # documentation for an example.
        result.append(
            get_partial_source(
                source, begin, begin + 2, marks=marks, nb_digits=nb_digits
            )
        )
        result.append(continuation)
        result.append(get_partial_source(source, end - 2, end, marks=marks))
    else:
        lines = source.split("\n")
        for index, line in enumerate(lines, start=1):
            if index < begin:
                continue
            if index > end:
                break
            if index in marks:
                result.append(with_mark.format(index) + line)
            else:
                result.append(no_mark.format(index) + line)

    return "\n".join(result)


def write_err(msg):
    """Writes a string to sys.stderr."""
    sys.stderr.write(msg)


def write_exception_info(exc, source):
    """Writes the information we have after processing the exception."""
    msg = handle_exception(exc, source)
    if msg is not None:
        write_err(msg)
    else:
        write_err(
            _("An exception was raised for which we have no simplified traceback:\n")
        )
        write_err("name: %s\n" % exc.__class__.__name__)
        write_err("args: " + str(exc.args) + "\n")
        write_err("Python traceback follows:\n\n")
        for item in traceback.format_tb(exc.__traceback__):
            write_err(item)
        write_err("\n")


def handle_exception(exc, source):
    """Generic function to handle exceptions and returns a
       friendlier traceback than Python.

       If the exception is not recognized as one for which a friendlier
       traceback can be provided, None is returned
    """
    if not ENABLED:
        # Let normal Python traceback through
        raise exc

    name = exc.__class__.__name__
    if name in dispatch:
        return dispatch[name](exc, source)
    else:
        return None  # just want to be explicit


# ================= Specific handlers below


def handle_IfNobreakError(exc, source):
    """Handles situation where ``nobreak`` was wrongly use as a
       replacement for ``else`` in an if/elif/else block.
    """
    params = exc.args[0]
    begin = params["if_linenumber"]
    end = params["linenumber"]
    marks = [end]
    partial_source = get_partial_source(source, begin, end, marks=marks)

    return _(
        """
    AVANTPY EXCEPTION: IfNobreakError

    Error found in file '{filename}' on line {nobreak_linenumber}.

    Dialect used: {dialect}

{partial_source}

    The AvantPy '{nobreak_kwd}' keyword cannot be used in
    an '{if_kwd}/{elif_kwd}/{else_kwd}' clause (Python: if/elif/else).

"""
    ).format(
        **{
            "filename": params["source_name"],
            "nobreak_kwd": params["nobreak keyword"],
            "partial_source": partial_source,
            "nobreak_linenumber": end,
            "dialect": params["dialect"],
            "if_kwd": params["if_kwd"],
            "elif_kwd": params["elif_kwd"],
            "else_kwd": params["else_kwd"],
        }
    )


def handle_MismatchedBracketsError(exc, source):
    """Handles situation where a bracket of one kind '([{'
       is closed with a bracket of a different kind.
    """
    params = exc.args[0]
    begin = params["open_linenumber"]
    end = params["close_linenumber"]
    marks = [begin, end]
    partial_source = get_partial_source(source, begin, end, marks=marks)

    return _(
        """
    AVANTPY EXCEPTION: MismatchedBracketsError

    Error found in file '{filename}' on lines [{begin} - {end}].

    Dialect used: {dialect}

{partial_source}

    The opening {open_bracket} does not match the closing {close_bracket}.

"""
    ).format(
        **{
            "filename": params["source_name"],
            "open_bracket": params["open_bracket"],
            "close_bracket": params["close_bracket"],
            "begin": begin,
            "end": end,
            "partial_source": partial_source,
            "dialect": params["dialect"],
        }
    )


def handle_MissingLeftBracketError(exc, source):
    """Handles situation where at bracket of one kind ),],}
       is closed without a corresponding open bracket.
    """
    params = exc.args[0]
    linenumber = params["linenumber"]
    begin = linenumber - 1
    end = linenumber + 1
    marks = [linenumber]
    partial_source = get_partial_source(source, begin, end, marks=marks)

    return _(
        """
    AVANTPY EXCEPTION: MissingLeftBracketError

    Error found in file '{filename}' on line {linenumber}.

    Dialect used: {dialect}

{partial_source}

    The closing {bracket} does not match anything.

"""
    ).format(
        **{
            "filename": params["source_name"],
            "partial_source": partial_source,
            "dialect": params["dialect"],
            "linenumber": linenumber,
            "bracket": params["bracket"],
        }
    )


def handle_MissingRepeatColonError(exc, source):
    """Handles situation where a statement beginning with repeat
       does not end with a colon.
    """
    params = exc.args[0]
    linenumber = params["linenumber"]
    begin = linenumber - 1
    end = linenumber + 1
    marks = [linenumber]
    partial_source = get_partial_source(source, begin, end, marks=marks)

    return _(
        """
    AVANTPY EXCEPTION: MissingRepeatColonError

    Error found in file '{filename}' on line {linenumber}.

    Dialect used: {dialect}

{partial_source}

    A statement beginning with the '{repeat_kwd}' keyword must be on a single
    line ending with a colon (:) that indicates the beginning of an indented
    block of code, with no other colon appearing on that line.

"""
    ).format(
        **{
            "filename": params["source_name"],
            "repeat_kwd": params["repeat_kwd"],
            "partial_source": partial_source,
            "linenumber": linenumber,
            "dialect": params["dialect"],
        }
    )


def handle_MissingRepeatError(exc, source):
    """Handles situation where either ``until`` or ``forever`` was
       used without being preceeded by ``repeat``.
    """
    params = exc.args[0]
    linenumber = params["linenumber"]
    begin = linenumber - 1
    end = linenumber + 1
    marks = [linenumber]
    partial_source = get_partial_source(source, begin, end, marks=marks)

    return _(
        """
    AVANTPY EXCEPTION: MissingRepeatError

    Error found in file '{filename}' on line {linenumber}.

    Dialect used: {dialect}

{partial_source}

    The AvantPy '{keyword}'' keyword can be used only when preceded by '{repeat_kwd}'.

"""
    ).format(
        **{
            "filename": params["source_name"],
            "keyword": params["keyword"],
            "partial_source": partial_source,
            "linenumber": linenumber,
            "dialect": params["dialect"],
            "repeat_kwd": params["repeat_kwd"],
        }
    )


def handle_NameError(exc, source):
    """Handles situation where a NameError is raise."""
    msg = exc.args[0]
    var_name = msg.split("'")[1]
    exc_name = exc.__class__.__name__
    python_display = "{exc_name}: {msg}".format(exc_name=exc_name, msg=msg)

    last_tb_line = traceback.format_tb(exc.__traceback__)[-1]
    # For last_tb_line, we expect something like
    # File "filename", line 1, in <module>
    splitted_line = last_tb_line.split(",")
    filename = last_tb_line[0].replace("File ", "").replace('"', "").strip()
    if filename == "<string>" and state.current_filename is not None:
        filename = state.current_filename

    linenumber = int(splitted_line[1].replace("line ", ""))
    begin = linenumber - 1
    end = linenumber + 1
    marks = [linenumber]

    partial_source = get_partial_source(source, begin, end, marks=marks)

    return _(
        """
    PYTHON EXCEPTION: {python_display}

    Error found in file '{filename}' on line {linenumber}.

    Dialect used: {dialect}

{partial_source}

    A NameError exception indicates that a variable or
    function name is not known to Python.
    Most often, this is because there is a spelling mistake; however,
    sometimes it is because it is used before being defined or given a value.
    In your program, the unknown name is '{var_name}'.

"""
    ).format(
        **{
            "filename": filename,
            "python_display": python_display,
            "var_name": var_name,
            "partial_source": partial_source,
            "linenumber": linenumber,
            "dialect": state.current_dialect,
        }
    )


def handle_NobreakFirstError(exc, source):
    """Handles situation where ``nobreak`` was wrongly use as a
       replacement for ``else`` in statements like
       var = x if condition else y.
    """
    params = exc.args[0]
    linenumber = params["linenumber"]
    begin = linenumber - 1
    end = linenumber + 1
    marks = [linenumber]
    partial_source = get_partial_source(source, begin, end, marks=marks)

    return _(
        """
    AVANTPY EXCEPTION: NobreakFirstError

    Error found in file '{filename}' on line {linenumber}.

    Dialect used: {dialect}

{partial_source}

    The AvantPy '{nobreak_kwd}' keyword can be used instead of '{else_kwd}'
    (Python: else) only when it begins a new statement in
    '{for_kwd}/{while_kwd}' loops (Python: for/while).

"""
    ).format(
        **{
            "filename": params["source_name"],
            "nobreak_kwd": params["nobreak keyword"],
            "partial_source": partial_source,
            "linenumber": linenumber,
            "dialect": params["dialect"],
            "for_kwd": params["for_kwd"],
            "while_kwd": params["while_kwd"],
            "else_kwd": params["else_kwd"],
        }
    )


def handle_NobreakSyntaxError(exc, source):
    """Handles situation where ``nobreak`` is used without a matching
       ``for`` or ``while`` loop and not already raised.
    """
    params = exc.args[0]
    linenumber = params["linenumber"]
    begin = linenumber - 1
    end = linenumber + 1
    marks = [linenumber]
    partial_source = get_partial_source(source, begin, end, marks=marks)

    return _(
        """
    AVANTPY EXCEPTION: NobreakSyntaxError

    Error found in file '{filename}' on line {linenumber}.

    Dialect used: {dialect}

{partial_source}

    The AvantPy '{nobreak_kwd}' keyword can only be used as a replacement
    of '{else_kwd}' (Python: else) with a matching '{for_kwd}' or
    '{while_kwd}' loop (Python: for/while).

"""
    ).format(
        **{
            "filename": params["source_name"],
            "nobreak_kwd": params["nobreak keyword"],
            "partial_source": partial_source,
            "linenumber": linenumber,
            "dialect": params["dialect"],
            "for_kwd": params["for_kwd"],
            "while_kwd": params["while_kwd"],
            "else_kwd": params["else_kwd"],
        }
    )


def handle_RepeatFirstError(exc, source):
    """Handles situation where ``repeat`` was uses wrongly as
       it did not appear as the first keyword of a given statement.
    """
    params = exc.args[0]
    linenumber = params["linenumber"]
    begin = linenumber - 1
    end = linenumber + 1
    marks = [linenumber]
    partial_source = get_partial_source(source, begin, end, marks=marks)

    return _(
        """
    AVANTPY EXCEPTION: RepeatFirstError

    Error found in file '{filename}' on line {linenumber}.

    Dialect used: {dialect}

{partial_source}

    The AvantPy '{repeat_kwd}' keyword can only be used to begin
    a new loop (Python: equivalent to 'for' or 'while' loop).

"""
    ).format(
        **{
            "filename": params["source_name"],
            "repeat_kwd": params["repeat keyword"],
            "partial_source": partial_source,
            "linenumber": linenumber,
            "dialect": params["dialect"],
        }
    )


def handle_TryNobreakError(exc, source):
    """Handles situation where ``nobreak`` was wrongly use as a
       replacement for ``else`` in an try/except/else/finally block.
    """
    params = exc.args[0]
    begin = params["try_linenumber"]
    end = params["linenumber"]
    marks = [end]
    partial_source = get_partial_source(source, begin, end, marks=marks)

    return _(
        """
    AVANTPY EXCEPTION: TryNobreakError

    Error found in file '{filename}' on line {nobreak_linenumber}.

    Dialect used: {dialect}

{partial_source}

    The AvantPy '{nobreak_kwd}' keyword cannot be used in
    a '{try_kwd}/{except_kwd}/{else_kwd}/{finally_kwd}' clause
    (Python: try/except/else/finally).

"""
    ).format(
        **{
            "filename": params["source_name"],
            "nobreak_kwd": params["nobreak keyword"],
            "partial_source": partial_source,
            "nobreak_linenumber": end,
            "dialect": params["dialect"],
            "try_kwd": params["try_kwd"],
            "except_kwd": params["except_kwd"],
            "else_kwd": params["else_kwd"],
            "finally_kwd": params["finally_kwd"],
        }
    )


def handle_UnknownDialectError(exc, *args):
    """Handles error raised when an unknown dialect is requested
    """
    dialect = exc.args[0]
    all_dialects = exc.args[1]
    return _(
        """
    AVANTPY EXCEPTION: UnknownDialectError

    The following unknown dialect was requested: {dialect}.

    The known dialects are: {all_dialects}.

"""
    ).format(dialect=dialect, all_dialects=all_dialects)


def handle_UnknownLanguageError(exc, *args):
    """Handles error raised when an unknown language is requested
    """
    lang = exc.args[0]
    all_langs = exc.args[1]

    return _(
        """
    AVANTPY EXCEPTION: UnknownLanguageError

    The following unknown language was requested: {lang}.

    The known languages are: {all_langs}.

"""
    ).format(lang=lang, all_langs=all_langs)


dispatch = {
    "IfNobreakError": handle_IfNobreakError,
    "NameError": handle_NameError,
    "MismatchedBracketsError": handle_MismatchedBracketsError,
    "MissingLeftBracketError": handle_MissingLeftBracketError,
    "MissingRepeatColonError": handle_MissingRepeatColonError,
    "MissingRepeatError": handle_MissingRepeatError,
    "NobreakFirstError": handle_NobreakFirstError,
    "NobreakSyntaxError": handle_NobreakSyntaxError,
    "RepeatFirstError": handle_RepeatFirstError,
    "TryNobreakError": handle_TryNobreakError,
    "UnknownLanguageError": handle_UnknownLanguageError,
    "UnknownDialectError": handle_UnknownDialectError,
}



def test_repeat_first(capsys):
    from . import raise_repeat_first

    if not hasattr(raise_repeat_first, "err"):
        info = str(capsys.readouterr())
        raise_repeat_first.err = info
    assert "RepeatFirstError" in raise_repeat_first.err



def test_if_nobreak(capsys):
    from . import raise_if_nobreak

    if not hasattr(raise_if_nobreak, "err"):
        info = str(capsys.readouterr())
        raise_if_nobreak.err = info
    assert "IfNobreakError" in raise_if_nobreak.err



def test_try_nobreak(capsys):
    from . import raise_try_nobreak

    if not hasattr(raise_try_nobreak, "err"):
        info = str(capsys.readouterr())
        raise_try_nobreak.err = info
    assert "TryNobreakError" in raise_try_nobreak.err


def test_try_nobreak2(capsys):
    from . import raise_try_nobreak2

    if not hasattr(raise_try_nobreak2, "err"):
        info = str(capsys.readouterr())
        raise_try_nobreak2.err = info
    assert "TryNobreakError" in raise_try_nobreak2.err



def test_nobreak_first(capsys):
    from . import raise_nobreak_first

    if not hasattr(raise_nobreak_first, "err"):
        info = str(capsys.readouterr())
        raise_nobreak_first.err = info
    assert "NobreakFirstError" in raise_nobreak_first.err



def test_missing_repeat(capsys):
    from . import raise_missing_repeat

    if not hasattr(raise_missing_repeat, "err"):
        info = str(capsys.readouterr())
        raise_missing_repeat.err = info
    assert "MissingRepeatError" in raise_missing_repeat.err


def test_missing_repeat2(capsys):
    from . import raise_missing_repeat2

    if not hasattr(raise_missing_repeat2, "err"):
        info = str(capsys.readouterr())
        raise_missing_repeat2.err = info
    assert "MissingRepeatError" in raise_missing_repeat2.err



def test_nobreak_syntax(capsys):
    from . import raise_nobreak_syntax

    if not hasattr(raise_nobreak_syntax, "err"):
        info = str(capsys.readouterr())
        raise_nobreak_syntax.err = info
    assert "NobreakSyntaxError" in raise_nobreak_syntax.err


def test_nobreak_syntax2(capsys):
    from . import raise_nobreak_syntax2

    if not hasattr(raise_nobreak_syntax2, "err"):
        info = str(capsys.readouterr())
        raise_nobreak_syntax2.err = info
    assert "NobreakSyntaxError" in raise_nobreak_syntax2.err



"""pyupper.py
You can use this file as a template.

Inside a file named xx.py, create a dictionaries copied from
the example below, and name it 'xx' instead of 'upper'.
The keys should be identical to those of this file;
the values should be translated into the target language.
As an example, we provide a new English Python dialect where
we use 'function' as a synonym for 'lambda'.
"""

pyupper = {
    "False": "FALSE",
    "None": "NONE",
    "True": "TRUE",
    "and": "AND",
    "as": "AS",
    "assert": "ASSERT",
    "async": "async",  # do not translate
    "await": "await",  # as these are not for beginners
    "break": "BREAK",
    "class": "CLASS",
    "continue": "CONTINUE",
    "def": "DEF",
    "del": "DEL",
    "elif": "ELIF",
    "else": ["ELSE", "NOBREAK"],  # Special case
    "except": "EXCEPT",
    "finally": "FINALLY",
    "for": "FOR",
    "from": "FROM",
    "global": "GLOBAL",
    "if": "IF",
    "import": "IMPORT",
    "in": "IN",
    "is": "IS",
    "lambda": "FUNCTION",  # Clearer for beginners
    "nonlocal": "NONLOCAL",
    "not": "NOT",
    "or": "OR",
    "pass": "PASS",
    "raise": "RAISE",
    "return": "RETURN",
    "try": "TRY",
    "while": "WHILE",
    "with": "WITH",
    "yield": "YIELD",
    # Special loop keywords
    "repeat": "REPEAT",
    "forever": "FOREVER",
    "until": "UNTIL",
    # a few builtins useful for beginners
    "input": "INPUT",
    "print": "PRINT",
    "range": "RANGE",
    "exit": "EXIT",  # useful for console
    # a well-known python expression
    '__name__ == "__main__"': "NOTIMPORTED",
}



# pylint: skip-file
from setuptools import setup, find_packages

with open("README.md", encoding="utf8") as f:
    README = f.read()

with open("avantpy/version.py") as f:
    version_ = f.read().split("=")[-1].strip()[1:-1]

setup(
    name="avantpy",
    version=version_,
    description="Python with training wheels: executable pseudocode in any language.",
    long_description=README,
    long_description_content_type="text/markdown",
    classifiers=[
        "Development Status :: 1 - Planning",
        "License :: OSI Approved :: MIT License",
        "Programming Language :: Python :: 3",
        "Programming Language :: Python :: 3.7",
        "Intended Audience :: Education",
        "Topic :: Education",
    ],
    url="https://github.com/aroberge/avantpy",
    author="André Roberge",
    author_email="Andre.Roberge@gmail.com",
    packages=find_packages(exclude=["dist", "build", "tools"]),
    zip_safe=False,
)



"""
This module contains all the necessary functions, classes and methods
to take a file written in a given dialect and convert it into Python.
In doing so, it can identify some mistakes and raise some custom
exceptions which can be used to inform users in a way that is less
intimidating than the usual Python tracebacks.

The following can definitely be skipped unless you are interested in
(most of) the gory details.

AvantPy uses Python's ``tokenize`` module to convert a
source into a sequence of tokens. A token might be the name
of a variable, an operator, a parenthesis, a string, etc.

AvantPy analyses these tokens,
replacing some written into a different dialect until all
are converted into standard Python tokens.  Then, these
are recombined into a string which is the source to be
executed.

Python's tokenize module includes a function, called
untokenize, which can be used to combine a series of tokens
into a valid program.  With a normal Python program, doing
something similar to::

    new_source = untokenize(tokenize(source))

would be such that executing ``new_source`` would be the same
as executing ``source``.  However, the spacing between tokens
would not necessarily be the same for both ``new_source``
and the original program.  For example, the original program
may include a line like::

    variable = function( argument )

which might be converted into::

    variable =function(argument)

This could make it more difficult to compare the original
code with the converted one, as it is possible to do
using one of the utilises provided with AvantPy,
or any "diff" program.
As a result, we do not use Python's
untokenize function, and explicitly keep track of spacing
between tokens.

To understand how the conversion process works, it is useful to review
all possible cases, from some of the most complex, ending
with the simplest ones.

A. ``nobreak``

AvantPy has an additional keyword, named ``nobreak`` in the
English dialect, which can be used in ``for`` or ``while``
loops instead of the standard ``else``, as in::

    while condition:
        # code
    nobreak:
        # code

However, ``nobreak`` cannot be used in an ``if/elif/else``
blocks to replace ``else``.

Furthermore, ``nobreak`` cannot be used instead of ``else``
in an assignment such as::

    a = 1 if True else 2

To identify if a program includes a ``nobreak`` keyword
mistakenly, every time we see a leading ``for``, ``while``,
``if`` or ``elif`` keyword (or their equivalent in a
given dialect), we note the indentation (column where the
first character is written) and the corresponding keyword.
A list containing these keywords is called ``blocks_with_else``
in this function.

Later, when we encounter a ``nobreak`` keyword at a given
indentation, we check to see if the last ``blocks_with_else``
keyword found at that same indentation was one for which
it made sense to use ``nobreak`` or not.  If it was a
loop, we simply replace ``nobreak`` by ``else``. If not,
we raise a custom exception which is handled elsewhere.

B. ``repeat``

In addition to the standard Python loops constructs, AvantPy
support four additional idioms::

    repeat forever:           # while True:
        pass
    repeat while condition:   # while condition:
        pass
    repeat until condition:   # while not condition:
        pass
    repeat n:                 # for some_var in range(n):
        pass

For this last case, ``n`` could be an expression, possibly
spanning multiple lines.

When we encounter the equivalent to the "repeat" keyword in
the selected dialect, we must make sure that it is the first
keyword occurring on a logical line; if not, we raise a
custom exception.

If ``repeat`` is the first keyword on a line, we set a flag
(repeat_loop) to True, preparing to look at the next token.

a) If the next token is one of ``forever``, ``until``, ``while``,
or their equivalent in the target dialect
(remember that including normal Python keywords in a program written
in a different dialect is allowed)
we can proceed with the rest in a straightforward manner.

b) if that is not the case, we set a different flag (repeat_n)
to True so that we can deal with the relevant for loop.

For this last case, the variable in the for loop is a dummy
variable; we must ensure that its name is chosen such that
it does not occur anywhere else in the source code.
This is accomplished using a method called
``get_unique_variable_names``.

C. ``nobreak`` and ``repeat``

A ``repeat`` loop is essentially a ``for`` or a ``while``
loop. As such, it could have an ``else`` clause which
has a clearer meaning if the keyword ``nobreak`` is used
instead.  Thus, just like we mentioned before, we also
keep track of where a leading ``repeat`` is used.

D. Direct translation

If a token does not match one of the cases described above,
we need to see if it is a term used in the dialect; if
so, we simply translate it into standard Python.

E. Brackets

In Python, brackets must always come in pairs: (...), [...],
{...}. In the course of processing the file, if we identify
brackets which are not paired correctly, an exception
is raised.

F. Remaining tokens

Any remaining token is left as is; it is assumed to be valid
Python.
"""
import os
import tokenize
from io import StringIO

from . import exceptions
from .session import state
from .utils import Token


# =================================
# Note: the strings in this module do not need to be translated.
# In particular, the exception 'messages' can essentially be taken
# as comments, since they will be properly processed (and translated)
# in another module.
# =================================


def get_unique_variable_names(source, repeat_kwd, all_count_names=[]):
    """Returns a list of possible variables names that
       are not found in the original source and have not been used
       anywhere.
       These variables will be used in expressions such as

           for _COUNT_42 in range(n):

       which will replace

           repeat n:
    """
    nb = source.count(repeat_kwd)
    if nb == 0:
        return []

    base_name = "_COUNT_"
    var_names = []
    i = 0
    j = 0
    while j < nb:
        tentative_name = base_name + str(i)
        if source.count(tentative_name) == 0 and tentative_name not in all_count_names:
            var_names.append(tentative_name)
            all_count_names.append(tentative_name)
            j += 1
        i += 1
    return var_names


class Converter:
    """This class uses Python's ``tokenize`` module to convert a
    source into a sequence of tokens. A token might be the name
    of a variable, an operator, a parenthesis, a string, etc.

    It processes one token at a time, identifying potential misuses,
    replacing some tokens written into a different dialect until all
    are converted into standard Python tokens.  Then, these
    are recombined into a string which is the source to be
    executed.

    Python's tokenize module includes a function, called
    untokenize, which can be used to combine a series of tokens
    into a valid program.  With a normal Python program, doing
    something similar to::

        new_source = untokenize(tokenize(source))

    would be such that executing ``new_source`` would be the same
    as executing ``source``.  However, the spacing between tokens
    would not necessarily be the same for both ``new_source``
    and the original program.  For example, the original program
    may include a line like::

        variable = function( argument )

    which might be converted into::

        variable =function(argument)

    This could make it more difficult to compare the original
    code with the converted one, as it is possible to do
    using one of the utilises provided with AvantPy,
    or any "diff" program.
    As a result, we do not use Python's
    untokenize function, and explicitly keep track of spacing
    between tokens."""

    def __init__(self, source, dialect=None, source_name=None):
        """Initializes variables that are used to convert a program
           written in a given dialect into Standard Python.
        """
        if dialect is not None:
            state.set_dialect(dialect)
        self.dialect = state.get_dialect()
        self.source = source

        if self.dialect is not None:
            self.source_name = source_name
            self.init_dialect_vars()
            self.init_bookkeeping_vars()

    def convert(self):
        """Uses Python's tokenize module to generate tokens from a source.

           It calls a function to process each token in turn, and combine
           the results to returned a transformed source.
        """
        if self.dialect is None:
            return self.source
        tokens = tokenize.generate_tokens(StringIO(self.source).readline)
        for tok in tokens:
            token = Token(tok)
            self.process_token(token)

        return "".join(self.result)

    def init_dialect_vars(self):
        """Initialises variables that are dependent on a given dialect."""

        self.lang_to_py = state.get_to_python(self.dialect)
        py_to_lang = state.get_from_python(self.dialect)

        self.if_kwd = py_to_lang["if"]
        self.elif_kwd = py_to_lang["elif"]
        self.else_kwd = py_to_lang["else"][0]
        self.nobreak_kwd = py_to_lang["else"][1]
        self.repeat_kwd = py_to_lang["repeat"]
        self.for_kwd = py_to_lang["for"]
        self.while_kwd = py_to_lang["while"]
        self.until_kwd = py_to_lang["until"]
        self.forever_kwd = py_to_lang["forever"]
        self.loops_with_else = [
            "for",
            "while",
            self.for_kwd,
            self.while_kwd,
            self.repeat_kwd,
        ]
        self.if_blocks = ["if", self.if_kwd]
        self.try_kwd = py_to_lang["try"]
        self.except_kwd = py_to_lang["except"]
        self.finally_kwd = py_to_lang["finally"]
        self.try_blocks = ["try", self.try_kwd]
        self.blocks_with_else = self.if_blocks + self.try_blocks + self.loops_with_else

    def init_bookkeeping_vars(self):
        """Initialises variables that are used for bookkeeping"""
        self.brackets = []  # keeps track where (), [], {} occur
        self.indentations = {}  # keeps track of indentation levels of some blocks

        # variable names to be used in
        #    for variable_name in range(...):
        self.count_names = get_unique_variable_names(self.source, self.repeat_kwd)

        self.result = []  # accumulates transformed tokens
        self.prev_lineno = -1
        self.prev_col = 0
        self.just_processed_repeat_kwd = False
        self.repeat_n = False
        self.repeat_line = False
        self.begin_new_line = True

    def process_token(self, token):
        """Determines what to do for each individual token."""

        # including indentation as a token can mess up the determination
        # of whether or not a keyword appears at the beginning of a new
        # line. Thus, we discard such spaces, and keep track of the required
        # spacing elsewhere.
        if not token.string.strip(" \t"):
            return

        self.begin_new_line = token.start_line != self.prev_lineno

        self.preserve_repeat_spacing(token)

        if token.string == self.repeat_kwd:
            self.process_repeat(token)

        elif self.just_processed_repeat_kwd:
            self.process_after_repeat(token)

        elif self.repeat_n and token.string == ":":
            if self.brackets:
                self.raise_missing_repeat_colon(token.start_line)
            self.result.append("):")
            self.repeat_line = False
            self.repeat_n = False

        elif self.begin_new_line and self.repeat_line:
            self.raise_missing_repeat_colon(token.start_line - 1)

        elif self.repeat_line and token.string == ":":
            if self.brackets:  # Inside a slice or other construct with colon
                self.raise_missing_repeat_colon(token.start_line)
            self.repeat_line = False
            self.result.append(":")

        elif not self.just_processed_repeat_kwd and token.string in [
            self.forever_kwd,
            self.until_kwd,
        ]:
            self.do_until_forever_error(token)

        elif token.string in self.blocks_with_else:
            if self.begin_new_line:
                # keeping track of beginning of for/while/if block so that
                # we can tell if replacing 'nobreak' by 'else' makes sense
                self.indentations[token.start_col] = [token.string, token.start_line]
            if token.string in self.lang_to_py:
                self.result.append(self.lang_to_py[token.string])
            else:
                self.result.append(token.string)

        elif token.string in "([{":
            self.brackets.append((token.string, token.start_line))
            self.result.append(token.string)

        elif token.string in ")]}":
            self.do_close_bracket(token)

        elif token.string == self.nobreak_kwd:
            self.process_nobreak(token)

        elif token.string in self.lang_to_py:
            self.result.append(self.lang_to_py[token.string])

        else:
            self.result.append(token.string)

    def raise_missing_repeat_colon(self, linenumber):
        """Raised if a repeat statement does not end with a colon on the same
           line with no other colon appearing on that line.
        """
        raise exceptions.MissingRepeatColonError(
            "Missing colon on line beginning with repeat",
            (
                {
                    "repeat_kwd": self.repeat_kwd,
                    "linenumber": linenumber,
                    "source_name": self.source_name,
                    "source": self.source,
                    "dialect": self.dialect,
                },
            ),
        )

    def do_close_bracket(self, token):
        """Determines if a closing bracket ), ], or }, matches a previously
           opened one, and raises an error if that is not the case.
        """
        try:
            previous_bracket = self.brackets.pop()
        except IndexError:
            raise exceptions.MissingLeftBracketError(
                "Closing bracket found with no matching opening one",
                (
                    {
                        "bracket": token.string,
                        "linenumber": token.start_line,
                        "source_name": self.source_name,
                        "source": self.source,
                        "dialect": self.dialect,
                    },
                ),
            )

        open_bracket = previous_bracket[0]
        if (
            (open_bracket == "(" and token.string != ")")
            or (open_bracket == "[" and token.string != "]")
            or (open_bracket == "{" and token.string != "}")
        ):
            raise exceptions.MismatchedBracketsError(
                "Closing bracket found matching a different opening one.",
                (
                    {
                        "close_bracket": token.string,
                        "open_bracket": open_bracket,
                        "close_linenumber": token.start_line,
                        "open_linenumber": previous_bracket[1],
                        "source_name": self.source_name,
                        "source": self.source,
                        "dialect": self.dialect,
                    },
                ),
            )
        else:
            self.result.append(token.string)

    def preserve_repeat_spacing(self, token):
        """ We ensure spacing of original file is preserved,
            including space between tokens.
            When self.just_processed_repeat_kwd is True, it means we just saw
            the repeat keyword which can simply be dropped;
            we don't need add the space between "repeat" and the next token.
        """
        if not self.just_processed_repeat_kwd:
            if token.start_line > self.prev_lineno:
                self.prev_col = 0
            if token.start_col > self.prev_col and token.string != "\n":
                self.result.append(" " * (token.start_col - self.prev_col))
        self.prev_col = token.end_col
        self.prev_lineno = token.end_line

    def do_until_forever_error(self, token):
        """This function is called when we see a keyword equivalent to
           ``until`` or ``forever`` in a given dialect without them
           being preceeded by the keyword ``repeat``. This is not allowed.
        """
        assert not self.just_processed_repeat_kwd
        raise exceptions.MissingRepeatError(
            "until and forever must be preceeded by repeat",
            (
                {
                    "keyword": token.string,
                    "linenumber": token.start_line,
                    "source_name": self.source_name,
                    "source": self.source,
                    "dialect": self.dialect,
                    "repeat_kwd": self.repeat_kwd,
                },
            ),
        )

    def process_after_repeat(self, token):
        """After we see the ``repeat`` keyword, there are four possibilities:

               1. the keyword ``until`` follows

               2. the keyword ``until`` follows

               3. the keyword ``forever`` follows

               4. it is meant to be followed by some_expression which evaluates
                  to an integer.
        """
        self.just_processed_repeat_kwd = False
        if token.string in [self.while_kwd, self.until_kwd, self.forever_kwd]:
            if token.string == self.while_kwd:
                self.result.append("while")
            elif token.string == self.until_kwd:
                self.result.append("while not")
            else:
                self.result.append("while True")
        else:
            self.repeat_n = True
            self.result.append("for %s in range(" % self.count_names.pop())
            self.result.append(token.string)

    def process_nobreak(self, token):
        """The ``nobreak`` keyword is allowed to replace ``else`` in
           a ``while`` or ``for`` loop only.
        """
        if (
            self.begin_new_line
            and token.start_col in self.indentations
            and self.indentations[token.start_col][0] in self.loops_with_else
        ):
            self.result.append("else")
        else:
            self.handle_nobreak_error(token)

    def process_repeat(self, token):
        """This identifies if ``repeat`` begins a new line; if so, we
           simply set up a flag to indicate that we just saw this keyword,
           and otherwise drop it.

           If it does not begin a new line, we raise an exception.
        """
        if not self.begin_new_line:  # this is not allowed to happen
            raise exceptions.RepeatFirstError(
                "repeat must be first statement on a line",
                (
                    {
                        "repeat keyword": token.string,
                        "linenumber": token.start_line,
                        "source_name": self.source_name,
                        "source": self.source,
                        "dialect": self.dialect,
                    },
                ),
            )
        self.just_processed_repeat_kwd = True
        self.repeat_line = True

    def handle_nobreak_error(self, token):
        """Deals with identified misuses of the ``nobreak`` keyword."""
        if not self.begin_new_line:
            raise exceptions.NobreakFirstError(
                "nobreak must be first statement on a line",
                (
                    {
                        "nobreak keyword": token.string,
                        "linenumber": token.start_line,
                        "source_name": self.source_name,
                        "source": self.source,
                        "dialect": self.dialect,
                        "for_kwd": self.for_kwd,
                        "while_kwd": self.while_kwd,
                        "else_kwd": self.else_kwd,
                    },
                ),
            )
        elif (
            token.start_col in self.indentations
            and self.indentations[token.start_col][0] in self.if_blocks
        ):
            raise exceptions.IfNobreakError(
                "Keyword nobreak found matching if/elif",
                (
                    {
                        "if_string": self.indentations[token.start_col][0],
                        "if_linenumber": self.indentations[token.start_col][1],
                        "nobreak keyword": token.string,
                        "linenumber": token.start_line,
                        "source_name": self.source_name,
                        "dialect": self.dialect,
                        "if_kwd": self.if_kwd,
                        "elif_kwd": self.elif_kwd,
                        "else_kwd": self.else_kwd,
                    },
                ),
            )
        elif (
            token.start_col in self.indentations
            and self.indentations[token.start_col][0] in self.try_blocks
        ):
            raise exceptions.TryNobreakError(
                "Keyword nobreak found matching try/except",
                (
                    {
                        "try_string": self.indentations[token.start_col][0],
                        "try_linenumber": self.indentations[token.start_col][1],
                        "nobreak keyword": token.string,
                        "linenumber": token.start_line,
                        "source_name": self.source_name,
                        "dialect": self.dialect,
                        "try_kwd": self.try_kwd,
                        "except_kwd": self.except_kwd,
                        "else_kwd": self.else_kwd,
                        "finally_kwd": self.finally_kwd,
                    },
                ),
            )
        else:
            raise exceptions.NobreakSyntaxError(
                "Keyword nobreak not matching a valid block",
                (
                    {
                        "nobreak keyword": token.string,
                        "linenumber": token.start_line,
                        "source_name": self.source_name,
                        "source": self.source,
                        "dialect": self.dialect,
                        "for_kwd": self.for_kwd,
                        "while_kwd": self.while_kwd,
                        "else_kwd": self.else_kwd,
                    },
                ),
            )


def convert(source, dialect=None, source_name=None):
    """Normally used from other functions to call for a conversion
       of a Python source from a given dialect into Python.
    """
    converter = Converter(source, dialect=dialect, source_name=source_name)
    return converter.convert()


def transcode(source, from_dialect, to_dialect):
    """Transforms a source (string) written in a known dialect into
       an equivalent version written in a different dialect.
       Spacing between tokens is preserved. This means that if a source
       is transformed from dialect xx to dialect yy, and that the result
       is transformed back to dialect xx, the original source should be
       recovered.

       Returns either the transformed source as a string, or None
       if something prevented the transformation to be successful.
    """
    # In many ways, this is a simplified version of Convert, but where
    # we assume that the syntax of the source is valid.
    if from_dialect == to_dialect:
        return source

    if to_dialect == "py":
        return convert(source, dialect=from_dialect, source_name="<transcode>")

    if from_dialect == "py":
        from_python = True
        from_dialect = "pyen"
    else:
        from_python = False

    from_lang = state.get_from_python(from_dialect)
    to_lang = state.get_from_python(to_dialect)

    if from_python:
        from_lang["lambda"] = "lambda"  # intead of "function" in pyen

    new_dict = {}
    for key, from_value in from_lang.items():
        to_value = to_lang[key]
        if isinstance(from_value, str):
            new_dict[from_value] = to_value
        else:
            for f, t in zip(from_value, to_value):
                new_dict[f] = t

    result = []  # accumulates transformed tokens
    prev_lineno = -1
    prev_col = 0

    tokens = tokenize.generate_tokens(StringIO(source).readline)
    for tok in tokens:
        token = Token(tok)
        if not token.string.strip(" \t"):
            continue
        if token.start_line > prev_lineno:
            prev_col = 0
        if token.start_col > prev_col and token.string != "\n":
            result.append(" " * (token.start_col - prev_col))
        prev_col = token.end_col
        prev_lineno = token.end_line

        if token.string in new_dict:
            result.append(new_dict[token.string])
        else:
            result.append(token.string)

    result = "".join(result)

    if from_python:
        result = "".join(result)
        result.replace('__name__ == "__main__"', to_lang['__name__ == "__main__"'])

    return result


def transcode_file(from_path, to_path):
    """Convenience function to be used from the command line."""

    cwd = os.getcwd()
    source_path = os.path.normpath(os.path.abspath(os.path.join(cwd, from_path)))
    if not os.path.isfile(source_path):
        print("%s does not exist." % source_path)
        return
    target_path = os.path.normpath(os.path.abspath(os.path.join(cwd, to_path)))

    _, extension = os.path.splitext(source_path)
    from_dialect = extension[1:]

    _, extension = os.path.splitext(target_path)
    to_dialect = extension[1:]

    with open(source_path, encoding="utf8") as f:
        source = f.read()

    target = transcode(source, from_dialect, to_dialect)

    with open(target_path, "w", encoding="utf8") as f:
        f.write(target)



def test_unknown_dialect(capsys):
    from . import raise_unknown_dialect

    if not hasattr(raise_unknown_dialect, "err"):
        info = str(capsys.readouterr())
        raise_unknown_dialect.err = info
    assert "UnknownDialectError" in raise_unknown_dialect.err



"""A custom importer making use of the import hook capability
"""
import importlib
import os
import os.path
import sys

from importlib.abc import Loader, MetaPathFinder
from importlib.util import spec_from_file_location

from . import session
from . import converter
from .exception_handling import write_exception_info
from .session import state

MAIN_MODULE_NAME = None


def import_main(name):
    """Imports the module that is to be interpreted as the main module.

       avantpy is often invoked with a script meant to be run as the
       main module its source is transformed with the -s (or --source) option,
       as in::

           python -m avantpy -s name

       Python identifies avantpy as the main script, which is not what we want.
    """
    global MAIN_MODULE_NAME
    MAIN_MODULE_NAME = name
    try:
        main = importlib.import_module(name)
        return main
    except ModuleNotFoundError:
        print(_("Cannot find main module: "), name)


class AvantPyMetaFinder(MetaPathFinder):
    """A custom finder to locate modules.  The main reason for this code
       is to ensure that our custom loader, which does the code transformations,
       is used."""

    def find_spec(self, fullname, path, target=None):
        """Finds the appropriate properties (spec) of a module, and sets
           its loader."""

        if not path:
            path = [os.getcwd()]
        if "." in fullname:
            name = fullname.split(".")[-1]
        else:
            name = fullname
        for entry in path:
            for ext in session.state.all_dialects():
                filename = os.path.join(entry, name + "." + ext)
                if os.path.exists(filename):
                    break
            else:
                continue

            return spec_from_file_location(
                fullname,
                filename,
                loader=AvantPyLoader(filename),
                submodule_search_locations=None,
            )
        return None  # Not an AvantPy file


sys.meta_path.insert(0, AvantPyMetaFinder())


class AvantPyLoader(Loader):
    """A custom loader which will transform the source prior to its execution"""

    def __init__(self, filename):
        self.filename = filename

    def exec_module(self, module):
        """import the source code, converts it before executing it."""

        if module.__name__ == MAIN_MODULE_NAME:
            module.__name__ = "__main__"

        with open(self.filename, encoding="utf8") as f:
            source = f.read()
        original = source

        _path, extension = os.path.splitext(self.filename)
        name = os.path.basename(_path)
        fullname = name + extension
        dialect = extension[1:]

        try:
            session.state.set_dialect(dialect)
            source = converter.convert(source, dialect, source_name=fullname)
        except Exception as exc:
            write_exception_info(exc, original)
            return

        # ------------------------
        # Ideally, instead of the following use of exec(source), we would
        # proceed in two steps:
        # 1. use a custom AST parser that could generate more detailed
        #    information when a SyntaxError is found
        # 2. If no error is found, exec the code objects produced by the AST.
        # -------------------------

        try:
            exec(source, vars(module))
        except Exception as exc:
            state.current_filename = fullname
            write_exception_info(exc, original)
